/*
	ProWesS, PROGS Windowing System

	PWCreate

	(c) PROGS, Professional & Graphical Software
	written by Joachim Van der Auwera
	started 14 January 1995
	18.01.1996 system is now allocated inside a Heap object, which is also used
			   for the keypresses and regions.
	16.01.1997 PWDoCreate added
*/

#include <mem.h>
#include <win.h>

#include "private.h"
#include "PWHandler.h"
#include "PWJob.h"

#ifndef catch
#define catch(x) do { if (err=(x)) return err; } while (0)
#endif

/* external symbols */
extern Sprite sprite_diagprim;
extern Sprite sprite_diagsec;

#define HEAP_BLOCKSIZE		2048

/* forward declaration */
static Error PWGetRegion(PWSystem *system, PWRegion **ret);
static void linkin_region(PWRegion *region, PWRegion *into);
static Error position_region(PWSystem *system, PWRegion *region,
							 PWRegion *owner, char position, char direction);
static Error inlevel_region(PWSystem *system, PWRegion *region, PWRegion *owner, char direction);
extern Bbox emptyBbox;
#define BEFORE	1
#define AFTER	2
#define ROW 	1
#define COLUMN	0

Error PWDoCreate(PWObject owner, PWObject *ret, PWType type, int *taglist)
{
	Error err;
	PWSystem *system;
	PWTypeDef *typedefn;
	char event_linked;
	PWEvent event;
	PWObject temp;
	int *orgtaglist=taglist;

	if (ret==NULL) ret=&temp;
	*ret=NULL;						/* in case of error */

	/* search for the typedefinition */

	catch( PWGetType(type,&typedefn) );

	/* get the system and job information (from owner) */

	if (owner==NULL)
	{
		Heap heap;

		/* create a new system for this job */
		catch( MEMHeapInit(HEAP_BLOCKSIZE,&heap) );
		err=MEMHeapAllocate(heap,sizeof(PWSystem),&system);
		if (err) { MEMHeapZap(heap); return err; }
		MEMClear(sizeof(PWSystem),system);

		/* and initialised */
		system->heap=heap;
		system->window=WINDOW_NONE;
		system->win_xpos=system->win_ypos=
		system->win_ptrx=system->win_ptry=-1;
		system->subwindows.list=system->subwinlist;
		system->subwinlist[0]=&system->subwin[0];
		system->subwinlist[1]=&system->subwin[1];
		system->subwinlist[2]=&system->subwin[2];
		system->subwinlist[3]=&system->subwin[3];
		system->subwinlist[4]=&system->subwin[4];
		system->subwinlist[5]=&system->subwin[5];
		system->subwinlist[6]=&system->subwin[6];
		system->subwinlist[7]=&system->subwin[7];
		system->subwinlist[8]=NULL;
		system->subwindows.main.sprite=NULL;
		system->subwin[0].sprite=
		system->subwin[1].sprite=
		system->subwin[6].sprite=
		system->subwin[7].sprite=&sprite_diagprim;
		system->subwin[2].sprite=
		system->subwin[3].sprite=
		system->subwin[4].sprite=
		system->subwin[5].sprite=&sprite_diagsec;
		catch( PFGstateInit(PF_DRIVER_DUMMY,NULL,&emptyBbox,&system->gstate,NULL) );
	}
	else
	{
		switch (owner->type->event_type)
		{
		case PW_TYPE_EVENT_REGION:
			system=owner->event.region->system;
			break;
		case PW_TYPE_EVENT_KEYPRESS:
			system=owner->event.keypress->system;
			break;
		}
	}

	
	/* create an event to match the type of object */
	switch(typedefn->event_type)
	{
	case PW_TYPE_EVENT_REGION:
		catch( PWGetRegion(system,&event.region) );

		event.region->system=system;
		event.region->parentObject=owner;

		event_linked=FALSE;
		break;
	case PW_TYPE_EVENT_KEYPRESS:
		/* allocate */
		catch( MEMHeapAllocate(system->heap,sizeof(PWKeyPress),&event.keypress) );
		MEMClear(sizeof(PWKeyPress),event.keypress);

		/* initialise and link in */
		event.keypress->system=system;
		event.keypress->parentObject=owner;
		event.keypress->next=system->keypresses;
		system->keypresses=event.keypress;
		event_linked=TRUE;
		break;
	}


	/* process argument list */
	while (*taglist)
	{
		char tag_recognized=TRUE;

		switch (*taglist)
		{
		case PW_POSITION_RIGHT_OF :
			if (event_linked) return ERR_IPAR;
			catch( position_region(system,event.region,
								((PWObject)(*(taglist+1)))->event.region,
								AFTER,ROW) );
			event_linked=TRUE;
			break;
		case PW_POSITION_LEFT_OF :
			if (event_linked) return ERR_IPAR;
			catch( position_region(system,event.region,
								((PWObject)(*(taglist+1)))->event.region,
								BEFORE,ROW) );
			event_linked=TRUE;
			break;
		case PW_POSITION_ABOVE :
			if (event_linked) return ERR_IPAR;
			catch( position_region(system,event.region,
								((PWObject)(*(taglist+1)))->event.region,
								BEFORE,COLUMN) );
			event_linked=TRUE;
			break;
		case PW_POSITION_BELOW :
			if (event_linked) return ERR_IPAR;
			catch( position_region(system,event.region,
								((PWObject)(*(taglist+1)))->event.region,
								AFTER,COLUMN) );
			event_linked=TRUE;
			break;
		case PW_POSITION_NEXT_ROW :
			if (event_linked) return ERR_IPAR;
			catch( inlevel_region(system,event.region,owner->event.region,ROW) );
			event_linked=TRUE;
			break;
		case PW_POSITION_NEXT_COLUMN :
			if (event_linked) return ERR_IPAR;
			catch( inlevel_region(system,event.region,owner->event.region,COLUMN) );
			event_linked=TRUE;
			break;
		case PW_SLEEP_OBJECT:
			if (system->sleep)
				PWRemove(system->sleep->owner);
			system->sleep=event.region;
			event_linked=TRUE;
			break;
		default :
			tag_recognized=FALSE;
			break;
		}

		if (tag_recognized)
		{
			*taglist&=PW_TAGMASK_SKIP;
			if (*taglist==NUL) *taglist=PW_TAGMASK_MULTIPLE(1);
		}

		taglist=PWSkipTag(taglist);
	}


	/* link in object (if not yet done) */
	if (!event_linked)
	{
		/* this is a region event (keypresses are linked when created) */
		/* the region has to be linked as last region inside the parent */

		/* start by making sure that the parent is also a region */
		if (owner)
		{
			if (owner->type->event_type!=PW_TYPE_EVENT_REGION)
				return ERR_IPAR;

			/* now link in where appropriate */
			linkin_region(event.region,owner->event.region);
		}
		else
		{
			/* no parent, then link in as first region */
			event.region->level=1;
			event.region->parent=NULL;
			system->regions=event.region;
			event.region->next=event.region->prev=event.region->children=NULL;
		}
	}


	/* call object creator */
	catch( typedefn->PWCreate(owner, ret, event, orgtaglist) );

	/* fill in part of the object - to releave object create from the burdon */
	(*ret)->type=typedefn;
	(*ret)->event=event;
	/* the others are clear, but may be set by object creator */

	/* fill in owner object in event */
	switch (typedefn->event_type)
	{
	case PW_TYPE_EVENT_REGION:
		event.region->owner=*ret;
		break;
	case PW_TYPE_EVENT_KEYPRESS:
		event.keypress->owner=*ret;
		break;
	}

	/* also set all the objects using object modifier (for tags which are */
	/* not specific to create routine) */
	return PWDoChange(*ret,orgtaglist);
}

/* default linking of a region into a list of regions */
static void linkin_region(PWRegion *tolink, PWRegion *owner)
{
	/* now search last region in parent */
	if (owner->children)
	{
		owner=owner->children;
		while(owner->next) owner=owner->next;
		while (owner->children && !owner->composite)
		{
			owner=owner->children;
			while(owner->next) owner=owner->next;
		}
		owner->next=tolink;
		tolink->prev=owner;
		tolink->parent=owner->parent;
	}
	else
	{
		owner->children=tolink;
		tolink->parent=owner;
		tolink->prev=NULL;
	}
	tolink->next=NULL;
	tolink->children=NULL;
	if (tolink->parent)
		tolink->level=tolink->parent->level+1;
	else
		tolink->level=1;
}

static void adjust_level(PWRegion *region)
{
	if (region->parent)
		region->level=region->parent->level+1;
	else
		region->level=1;
	for (region=region->children;region;region=region->next)
		adjust_level(region);
}

static Error position_region(PWSystem *system, PWRegion *region,
							 PWRegion *owner, char position, char direction)
{
	/* see if owner is inside a region with the correct primary direction */
	if ((owner->level&1)==direction)
	{
		/* link in before or after the owner */
		region->parent=owner->parent;
		region->children=NULL;
		region->level=owner->level;
		if (position==BEFORE)
		{
			region->prev=owner->prev;
			region->next=owner;
			if (owner->prev)
				owner->prev->next=region;
			else
				if (owner->parent)
					owner->parent->children=region;
				else
					system->regions=region;
			owner->prev=region;
		}
		else
		{
			region->prev=owner;
			region->next=owner->next;
			if (owner->next)
				owner->next->prev=region;
			owner->next=region;
		}
	}
	else
	{
		PWRegion *container;
		Error err;

		/* we have to allocate a new container region which should contain */
		/* the owner and the region positioned relative with the owner */

		err=PWGetRegion(system,&container);

		/* replace owner by container - linkages */
		container->parent=owner->parent;
		container->next=owner->next;
		container->prev=owner->prev;
		container->children=owner;
		container->level=owner->level;
		if (container->next)
			container->next->prev=container;
		if (container->prev)
			container->prev->next=container;
		else
			if (container->parent)
				container->parent->children=container;
			else
				system->regions=container;

		/* make sure the container has the correct position */
		container->tot_xpos=owner->tot_xpos;
		container->tot_ypos=owner->tot_ypos;

		/* put owner in container */
		owner->prev=owner->next=NULL;
		owner->parent=container;
		if (owner->parent)
			owner->level=owner->parent->level+1;
		else
			owner->level=1;

		/* make sure the children of the owner still behave correctly */
		if (owner->children)
		{
			PWRegion *subcontainer;

			/* we have to allocate a new container region which should contain */
			/* the owner and the region positioned relative with the owner */
			catch( PWGetRegion(system,&subcontainer) );

			/* link in as appropriate */
			subcontainer->parent=owner;
			subcontainer->children=owner->children;
			owner->children=subcontainer;

			/* adjust the level indicators of the children */
			adjust_level(subcontainer);
		}

		/* link in region relative with owner */
		return position_region(system,region,owner,position,direction);
	}
	return ERR_OK;
}

static Error inlevel_region(PWSystem *system, PWRegion *region, PWRegion *owner, char direction)
{
	PWRegion *container, *trail;
	Error err;

	/* allocate a new container region to act as new row/column */
	catch( PWGetRegion(system,&container) );

	/* link in as appropriate */
	container->parent=owner;
	container->children=NULL;
	container->level=owner->level+1;
	if (owner->children)
	{
		for (trail=owner->children;trail->next;trail=trail->next);
		trail->next=container;
		container->prev=trail;
	}
	else
	{
		owner->children=container;
	}

	/* make sure the direction is correct */
	if (((container->level+1)&1)==direction)
	{
		container->children=region;
		region->parent=container;
		region->level=container->level+1;
	}
	else
	{
		/* wrong direction, need another level */
		inlevel_region(system,region,owner,direction);
	}

	return ERR_OK;
}

static Error PWGetRegion(PWSystem *system, PWRegion **ret)
{
	Error err;

	catch( MEMHeapAllocate(system->heap,sizeof(PWRegion),ret) );

	/* initialise region */
	MEMClear(sizeof(PWRegion),*ret);
	(*ret)->timeout=FOREVER;
	(*ret)->windowfit=TRUE;
	(*ret)->scale=1;

	system->status|=SSTATUS_WINCHANGE;
	return err;
}

Error PWCreate(PWObject owner, PWObject *ret, PWType type, ...)
{
	return PWDoCreate(owner,ret,type,(int *)((&type)+1));
}
